use cgmath::Point3;
use crossbeam::{self, channel::Receiver};
use rapier2d::{
	dynamics::{RigidBodyBuilder, RigidBodyHandle},
	geometry::{ColliderBuilder, ColliderHandle, CollisionEvent},
	na::vector,
	pipeline::ChannelEventCollector,
};
use std::{collections::HashMap, time::Instant};
use winit::event::{ElementState, MouseButton, MouseScrollDelta, TouchPhase, VirtualKeyCode};

use super::{camera::Camera, physics::Physics, ship, system::System, util, InputStatus};
use crate::{
	consts,
	content::Content,
	render::{Sprite, SpriteTexture},
};

pub struct ProjectileBuilder {
	pub rigid_body: RigidBodyBuilder,
	pub collider: ColliderBuilder,
	pub sprite: SpriteTexture,
	pub lifetime: f32,
	pub size: f32,
}

#[derive(Debug)]
pub struct Projectile {
	pub rigid_body: RigidBodyHandle,
	pub sprite: SpriteTexture,
	pub lifetime: f32,
	pub size: f32,
}

impl Projectile {
	pub fn tick(&mut self, t: f32) {
		self.lifetime -= t;
	}

	pub fn is_expired(&self) -> bool {
		return self.lifetime < 0.0;
	}
}

pub struct Game {
	pub input: InputStatus,
	pub last_update: Instant,
	pub player: ship::Ship,
	pub test: ship::Ship,
	pub system: System,
	pub camera: Camera,
	paused: bool,
	pub time_scale: f32,

	pub projectiles: HashMap<ColliderHandle, Projectile>,

	collision_handler: ChannelEventCollector,
	physics: Physics,

	collision_recv: Receiver<CollisionEvent>,
}

impl Game {
	pub fn new(ct: Content) -> Self {
		let r1 = RigidBodyBuilder::dynamic()
			.translation(vector![0.0, 0.0])
			.can_sleep(false)
			.build();
		let c1 = ColliderBuilder::ball(50.0)
			.restitution(0.7)
			.mass(1.0)
			.build();

		let r2 = RigidBodyBuilder::dynamic()
			.translation(vector![200.0, 200.0])
			.build();
		let c2 = ColliderBuilder::ball(50.0)
			.restitution(0.7)
			.mass(100.0)
			.build();

		let mut p = Physics::new();

		let rh1 = p.rigid_body_set.insert(r1);
		p.collider_set
			.insert_with_parent(c1, rh1, &mut p.rigid_body_set);

		let rh2 = p.rigid_body_set.insert(r2);
		p.collider_set
			.insert_with_parent(c2, rh2, &mut p.rigid_body_set);

		let (collision_send, collision_recv) = crossbeam::channel::unbounded();
		let (contact_force_send, _) = crossbeam::channel::unbounded();

		Game {
			last_update: Instant::now(),
			input: InputStatus::new(),
			projectiles: HashMap::new(),
			player: ship::Ship::new(
				&ct.ships[0],
				vec![
					ship::ShipOutfit::Gun(ship::ShipGun::new(ct.guns[0].clone(), 1)),
					ship::ShipOutfit::Gun(ship::ShipGun::new(ct.guns[0].clone(), 2)),
					ship::ShipOutfit::Engine(ct.engines[0].clone()),
				],
				rh1,
			),
			camera: Camera {
				pos: (0.0, 0.0).into(),
				zoom: 500.0,
			},
			system: System::new(&ct.systems[0]),

			paused: false,
			time_scale: 1.0,
			test: ship::Ship::new(&ct.ships[0], vec![], rh2),

			physics: p,
			collision_handler: ChannelEventCollector::new(collision_send, contact_force_send),
			collision_recv,
		}
	}

	pub fn process_key(&mut self, state: &ElementState, key: &VirtualKeyCode) {
		self.input.process_key(state, key)
	}

	pub fn process_click(&mut self, state: &ElementState, key: &MouseButton) {
		self.input.process_click(state, key)
	}

	pub fn process_scroll(&mut self, delta: &MouseScrollDelta, phase: &TouchPhase) {
		self.input.process_scroll(delta, phase)
	}

	pub fn set_paused(&mut self, pause: bool) {
		if pause {
			self.paused = true;
			self.input.release_all()
		} else {
			self.paused = false;
		}
	}

	pub fn update(&mut self) {
		let t: f32 = self.last_update.elapsed().as_secs_f32() * self.time_scale;

		self.physics.step(t, &self.collision_handler);

		self.projectiles.retain(|_, p| {
			p.tick(t);
			!p.is_expired()
		});

		// Update player and handle result
		self.player.update_controls(&self.input);
		let p = self.player.tick(&mut self.physics.rigid_body_set, t);
		for p in p.projectiles {
			let r = self.physics.rigid_body_set.insert(p.rigid_body.build());
			let c = self.physics.collider_set.insert_with_parent(
				p.collider.build(),
				r,
				&mut self.physics.rigid_body_set,
			);
			self.projectiles.insert(
				c,
				Projectile {
					rigid_body: r,
					sprite: p.sprite,
					lifetime: p.lifetime,
					size: p.size,
				},
			);
		}

		while let Ok(event) = self.collision_recv.try_recv() {
			if event.started() {
				self.projectiles.remove(&event.collider1());
				self.projectiles.remove(&event.collider2());
				// TODO: remove collider and rigid body,
				// cleanup, keep track of ships.
			}
		}

		if self.input.v_scroll != 0.0 {
			self.camera.zoom =
				(self.camera.zoom + self.input.v_scroll).clamp(consts::ZOOM_MIN, consts::ZOOM_MAX);
			self.input.v_scroll = 0.0;
		}

		// TODO: Camera physics
		let r = &mut self.physics.rigid_body_set[self.player.rigid_body];
		let ship_pos = util::rigidbody_position(r);
		self.camera.pos = ship_pos;

		self.last_update = Instant::now();
	}

	pub fn get_sprites(&self) -> Vec<Sprite> {
		let mut sprites: Vec<Sprite> = Vec::new();

		sprites.append(&mut self.system.get_sprites());
		sprites.push(self.player.get_sprite(&self.physics.rigid_body_set));
		sprites.push(self.test.get_sprite(&self.physics.rigid_body_set));

		// Make sure sprites are drawn in the correct order
		// (note the reversed a, b in the comparator)
		//
		// TODO: maybe use a gpu depth buffer instead?
		// I've tried this, but it doesn't seem to work with transparent textures.
		sprites.sort_by(|a, b| b.pos.z.total_cmp(&a.pos.z));

		// Don't waste time sorting these, they should always be on top.
		for (_, p) in &self.projectiles {
			let r = &self.physics.rigid_body_set[p.rigid_body];
			let pos = util::rigidbody_position(r);
			let ang = util::rigidbody_angle(r);
			sprites.push(Sprite {
				texture: p.sprite.clone(),
				pos: Point3 {
					x: pos.x,
					y: pos.y,
					z: 1.0,
				},
				size: p.size,
				angle: ang,
				children: None,
			})
		}

		return sprites;
	}
}
