use cgmath::{Deg, EuclideanSpace, Matrix2, Rad, Vector2};
use nalgebra::vector;
use rand::Rng;
use rapier2d::dynamics::{RigidBodyBuilder, RigidBodyHandle, RigidBodySet};
use rapier2d::geometry::ColliderBuilder;
use rapier2d::pipeline::ActiveEvents;

use super::super::util;
use super::{outfits::ShipOutfits, InputStatus, ShipControls, ShipOutfit, ShipTickResult};
use crate::game::game::ProjectileBuilder;
use crate::{
	content,
	render::{Sprite, SpriteTexture},
};

pub struct Ship {
	pub rigid_body: RigidBodyHandle,
	pub controls: ShipControls,
	outfits: ShipOutfits,

	sprite: SpriteTexture,
	size: f32,
}

impl Ship {
	pub fn new(ct: &content::Ship, outfits: Vec<ShipOutfit>, rigid_body: RigidBodyHandle) -> Self {
		let mut o = ShipOutfits::new(ct.engines.clone(), ct.guns.clone());
		for x in outfits.into_iter() {
			o.add(x)
		}

		Ship {
			rigid_body,
			controls: ShipControls::new(),

			outfits: o,
			sprite: SpriteTexture(ct.sprite.clone()),
			size: ct.size,
		}
	}

	pub fn update_controls(&mut self, input: &InputStatus) {
		self.controls.thrust = input.key_thrust;
		self.controls.right = input.key_right;
		self.controls.left = input.key_left;
		self.controls.guns = input.key_guns;
	}

	pub fn fire_guns(&mut self, rigid_body_set: &RigidBodySet) -> Vec<ProjectileBuilder> {
		let r = &rigid_body_set[self.rigid_body];
		let mut rng = rand::thread_rng();
		let mut out = Vec::new();

		for (g, p) in self.outfits.iter_guns_points() {
			if g.cooldown > 0.0 {
				continue;
			}

			g.cooldown = g.kind.rate + rng.gen_range(-g.kind.rate_rng..=g.kind.rate_rng);

			let ship_pos = util::rigidbody_position(r);
			let ship_ang: Deg<f32> = util::rigidbody_angle(r);
			let ship_ang_rad: Rad<f32> = ship_ang.into();
			let ship_vel = util::rigidbody_velocity(r);

			let pos = ship_pos + (Matrix2::from_angle(ship_ang) * p.pos.to_vec());

			let vel = ship_vel
				+ (Matrix2::from_angle(
					ship_ang + Deg(rng.gen_range(-(g.kind.spread.0 / 2.0)..=g.kind.spread.0 / 2.0)),
				) * Vector2 {
					x: 0.0,
					y: g.kind.projectile.speed
						+ rng.gen_range(-g.kind.projectile.speed_rng..=g.kind.projectile.speed_rng),
				});

			let p_r = RigidBodyBuilder::kinematic_velocity_based()
				.translation(vector![pos.x, pos.y])
				.rotation(-ship_ang_rad.0)
				.linvel(vector![vel.x, vel.y]);

			let p_c = ColliderBuilder::ball(5.0)
				.sensor(true)
				.active_events(ActiveEvents::COLLISION_EVENTS);

			out.push(ProjectileBuilder {
				rigid_body: p_r,
				collider: p_c,
				sprite: SpriteTexture(g.kind.projectile.sprite.clone()),
				lifetime: g.kind.projectile.lifetime
					+ rng.gen_range(
						-g.kind.projectile.lifetime_rng..=g.kind.projectile.lifetime_rng,
					),
				size: g.kind.projectile.size
					+ rng.gen_range(-g.kind.projectile.size_rng..=g.kind.projectile.size_rng),
			})
		}
		return out;
	}

	pub fn tick(&mut self, rigid_body_set: &mut RigidBodySet, t: f32) -> ShipTickResult {
		let r = &mut rigid_body_set[self.rigid_body];
		let ship_ang = util::rigidbody_angle(r);

		let engine_force = Matrix2::from_angle(ship_ang) * Vector2 { x: 0.0, y: 1.0 } * t;
		if self.controls.thrust {
			for e in self.outfits.iter_engines() {
				r.apply_impulse(vector![engine_force.x, engine_force.y] * e.thrust, true);
			}
		}

		if self.controls.right {
			r.apply_torque_impulse(500.0 * t, true);
		}

		if self.controls.left {
			r.apply_torque_impulse(-500.0 * t, true);
		}

		let p = if self.controls.guns {
			self.fire_guns(rigid_body_set)
		} else {
			Vec::new()
		};

		for i in self.outfits.iter_guns() {
			i.cooldown -= t;
		}

		return ShipTickResult { projectiles: p };
	}

	pub fn get_sprite(&self, rigid_body_set: &RigidBodySet) -> Sprite {
		let r = &rigid_body_set[self.rigid_body];
		let ship_pos = util::rigidbody_position(r);
		let ship_ang = util::rigidbody_angle(r);

		Sprite {
			pos: (ship_pos.x, ship_pos.y, 1.0).into(),
			texture: self.sprite.clone(), // TODO: sprite texture should be easy to clone
			angle: ship_ang,
			size: self.size,

			children: if self.controls.thrust {
				Some(self.outfits.get_engine_flares())
			} else {
				None
			},
		}
	}
}
